class PluginNotFound(Exception):
    pass


class OperationNotFound(Exception):
    pass


class TransformationNotFound(Exception):
    pass


class ElementNotFound(Exception):
    pass


class DuplicatedFeature(Exception):
    pass
